import torch
import torch.nn as nn
import math
import copy


def swish(x):
    return x * torch.sigmoid(x)


ACT2FN = {
    "gelu": torch.nn.functional.gelu,
    "relu": torch.nn.functional.relu,
    "swish": swish,
}


class Attention(nn.Module):
    def __init__(self, hidden_size, attention_head_num, dropout, vis=False):
        super().__init__()
        self.vis = vis
        self.attention_head_num = attention_head_num
        self.attention_head_size = int(hidden_size / self.attention_head_num)
        self.all_head_size = self.attention_head_num * self.attention_head_size

        self.query = nn.Linear(hidden_size, self.all_head_size)
        self.key = nn.Linear(hidden_size, self.all_head_size)
        self.value = nn.Linear(hidden_size, self.all_head_size)

        self.out = nn.Linear(hidden_size, hidden_size)
        self.attn_dropout = nn.Dropout(dropout)
        self.proj_dropout = nn.Dropout(dropout)

        self.softmax = nn.Softmax(dim=-1)

    def transpose_for_scores(self, x):
        new_x_shape = x.size()[:-1] + (
            self.attention_head_num,
            self.attention_head_size,
        )
        x = x.view(*new_x_shape)
        return x.permute(0, 2, 1, 3)

    def forward(self, hidden_states, attn_embedding=None):
        """attn_embedding: [1,Head,N,N]"""
        mixed_query_layer = self.query(hidden_states)
        mixed_key_layer = self.key(hidden_states)
        mixed_value_layer = self.value(hidden_states)

        query_layer = self.transpose_for_scores(mixed_query_layer)
        key_layer = self.transpose_for_scores(mixed_key_layer)
        value_layer = self.transpose_for_scores(mixed_value_layer)
        attention_scores = torch.matmul(query_layer, key_layer.transpose(-1, -2))
        attention_scores = attention_scores / math.sqrt(self.attention_head_size)
        if attn_embedding is not None:
            attention_scores = attention_scores + attn_embedding
        attention_probs = self.softmax(attention_scores)
        weights = attention_probs if self.vis else None
        attention_probs = self.attn_dropout(attention_probs)

        context_layer = torch.matmul(attention_probs, value_layer)
        context_layer = context_layer.permute(0, 2, 1, 3).contiguous()
        new_context_layer_shape = context_layer.size()[:-2] + (self.all_head_size,)
        context_layer = context_layer.view(*new_context_layer_shape)
        attention_output = self.out(context_layer)
        attention_output = self.proj_dropout(attention_output)
        return attention_output, weights


class Mlp(nn.Module):
    def __init__(self, hidden_size, dropout, mlp_dim):
        super(Mlp, self).__init__()
        self.fc1 = nn.Linear(hidden_size, mlp_dim)
        self.fc2 = nn.Linear(mlp_dim, hidden_size)
        self.act_fn = ACT2FN["gelu"]
        self.dropout = nn.Dropout(dropout)

        self._init_weights()

    def _init_weights(self):
        nn.init.xavier_uniform_(self.fc1.weight)
        nn.init.xavier_uniform_(self.fc2.weight)
        nn.init.normal_(self.fc1.bias, std=1e-6)
        nn.init.normal_(self.fc2.bias, std=1e-6)

    def forward(self, x):
        x = self.fc1(x)
        x = self.act_fn(x)
        x = self.dropout(x)
        x = self.fc2(x)
        x = self.dropout(x)
        return x


class Block(nn.Module):
    def __init__(
        self,
        hidden_size,
        mlp_dropout,
        attention_head_num,
        attention_dropout,
        vis=False,
        mlp_ratio=4,
    ):
        super(Block, self).__init__()
        self.hidden_size = hidden_size
        self.attention_norm = nn.LayerNorm(hidden_size, eps=1e-6)
        self.ffn_norm = nn.LayerNorm(hidden_size, eps=1e-6)
        self.ffn = Mlp(hidden_size, mlp_dropout, int(hidden_size * mlp_ratio))
        self.attn = Attention(hidden_size, attention_head_num, attention_dropout, vis)

    def forward(self, x, attn_embedding=None):
        h = x
        x = self.attention_norm(x)
        x, weights = self.attn(x, attn_embedding)
        x = x + h

        h = x
        x = self.ffn_norm(x)
        x = self.ffn(x)
        x = x + h
        return x, weights


class Encoder(nn.Module):
    def __init__(
        self,
        hidden_size,
        num_layers,
        mlp_dropout,
        attention_head_num,
        attention_dropout,
        mlp_ratio=4,
        vis=False,
    ):
        super().__init__()

        self.vis = vis
        self.layer = nn.ModuleList()
        self.encoder_norm = nn.LayerNorm(hidden_size, eps=1e-6)
        for _ in range(num_layers):
            layer = Block(
                hidden_size=hidden_size,
                mlp_dropout=mlp_dropout,
                attention_head_num=attention_head_num,
                attention_dropout=attention_dropout,
                vis=vis,
                mlp_ratio=mlp_ratio,
            )
            self.layer.append(copy.deepcopy(layer))

    def forward(self, hidden_states, attn_embedding=None):
        """attn_embedding: [1,Head,N,N]"""
        attn_weights = []
        for layer_block in self.layer:
            hidden_states, weights = layer_block(hidden_states, attn_embedding)
            if self.vis:
                attn_weights.append(weights)

        encoded = self.encoder_norm(hidden_states)
        return encoded, attn_weights



import argparse
import os.path as osp
import shutil
import sys
import pytorch_lightning as pl
from pytorch_lightning import seed_everything
from pytorch_lightning.callbacks import LearningRateMonitor, ModelCheckpoint
from pytorch_lightning.plugins import DDPPlugin
from pytorch_lightning.trainer import Trainer

from ..io.io import mkdir
from ..system import get_cpu_info, get_gpu_info, get_package_info
from ..tools import Logger, get_time_str
from .config import (get_log_dir, instantiate_from_config, load_config,
                     save_config)
from .dataset import LightningDataset
from .model import LightningModel
from .train_tips import get_train_tips

def train(config_filename: str) -> None:

    root_path = osp.abspath(config_filename).split("configs")[0]
    sys.path.insert(0, root_path)
    Logger.info(f"sys.path.insert(0, {root_path})")

    config = load_config(config_filename)

    task_name = config_filename.split("/")[1]
    exp_name = config_filename.split("/")[-1].split(".")[0]

    # log_dir
    log_dir = get_log_dir(config_filename)
    if osp.exists(log_dir):
        Logger.error("please create a new config!")
        return
    mkdir(log_dir)
    Logger.logfile(osp.join(log_dir, "train.log"), clear=True)
    Logger.info(f"mkdir: {log_dir}")

    # copy file
    copy_file_list = config.copy_file_list
    if copy_file_list is not None:
        for filename in copy_file_list:
            full_filename = osp.join(root_path, filename)
            if osp.exists(full_filename):
                dst = osp.join(log_dir, "code", filename)
                mkdir(osp.dirname(dst))
                shutil.copy(src=full_filename, dst=dst)
            else:
                Logger.warn(f"not exists: {full_filename}")

    # config
    new_config = osp.join(log_dir, "config.yaml")
    save_config(new_config, config)

    # log file
    Logger.info(f"train: {get_time_str()}")
    Logger.info(f"config: {config_filename}")

    # gpu, cpu, package info
    for key, value in get_gpu_info().items():
        Logger.info(f"{key}: {value}")
    for key, value in get_cpu_info().items():
        Logger.info(f"{key}: {value}")
    for key, value in get_package_info().items():
        Logger.info(f"{key}: {value}")

    # seed
    seed_everything(config.seed)
    Logger.info(f"seed:{config.seed}")

    Logger.info(f"gpus:{config.gpus}")
    Logger.info(f"precision:{config.precision}")
    Logger.info(f"train batch_size:{config.data.train.data_loader.batch_size}")
    Logger.info(f"train num_workers:{config.data.train.data_loader.num_workers}")
    Logger.info(f"save_top_k: {config.checkpoint.save_top_k}")
    Logger.info(f"save_last: {config.checkpoint.save_last}")

    resume_from_checkpoint = config.checkpoint.resume_from
    if resume_from_checkpoint:
        Logger.warn(f"resume from:{resume_from_checkpoint}")
    else:
        resume_from_checkpoint = None

    trainer = Trainer(
        precision=config.precision,
        gpus=config.gpus,
        sync_batchnorm=True,
        check_val_every_n_epoch=config.check_val_every_n_epoch,
        accumulate_grad_batches=config.accumulate_grad_batches,
        max_epochs=config.max_epochs,
        logger=[
            pl.loggers.TensorBoardLogger(
                save_dir="logs",
                name=task_name,
                version=exp_name,
            ),
        ],
        callbacks=[
            LearningRateMonitor(logging_interval="step"),
            ModelCheckpoint(
                dirpath=f"logs/{task_name}/{exp_name}/checkpoint",
                monitor="eval_loss",
                filename="{epoch:03d}_{eval_loss:.5f}",
                save_last=config.checkpoint.save_last,
                save_top_k=config.checkpoint.save_top_k,
            ),
        ],
        plugins=DDPPlugin(find_unused_parameters=config.find_unused_parameters),
        resume_from_checkpoint=resume_from_checkpoint,
        log_every_n_steps=config.log_every_n_steps,
        profiler="simple",
    )

    Logger.info("prepare to build model...")
    model = instantiate_from_config(config.model)
    lightning_model = LightningModel(
        model=model, optimizer_cfg=config.optimizer, scheduler_cfg=config.scheduler
    )

    Logger.info("prepare to build dataset...")
    data = LightningDataset(**config.data)
    
    Logger.info("prepare to train...")
    Logger.warn(f"tensorboard --logdir=logs/{task_name}")
    trainer.fit(lightning_model, data)


def train_main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "config", default="config.yaml", type=str, help="config filename"
    )

    args = parser.parse_args()

    if args.config == "tips":
        for cnt, tip in enumerate(get_train_tips()):
            title = tip["title"]
            url = tip["url"]
            Logger.warn(f"Tip {cnt+1}:")
            Logger.info(f"title: {title}")
            Logger.info(f"url: {url}")
            print()
        return
    elif not osp.exists(args.config):
        Logger.error(f"not exists: {args.config}")
        sys.exit(0)

    train(args.config)



import numpy as np
import pytorch_lightning as pl
import torch
from torch.utils.data import DataLoader
from torch.utils.data import Dataset
from ..tools.logger import Logger
from .config import instantiate_from_config
import sys


class RandomSplitDataset(Dataset):
    def __init__(self, dataset: Dataset, random_split: float = 1.0) -> None:
        super().__init__()

        self.dataset = dataset
        self.random_split = random_split

        N = len(self.dataset)
        n = int(N * float(random_split))
        self.__index_list = np.random.permutation(N)[:n].tolist()

    def __len__(self):
        return len(self.__index_list)

    def __getitem__(self, index):
        real_index = self.__index_list[index]
        return self.dataset.__getitem__(real_index)


class LightningDataset(pl.LightningDataModule):
    def __init__(self, train=None, validation=None, test=None):
        super().__init__()
        self.configs = {
            "train": train,
            "validation": validation,
            "test": test,
        }
        if train is not None:
            self.train_dataloader = self._train_dataloader
        if validation is not None:
            self.val_dataloader = self.val_dataloader
        if test is not None:
            self.test_dataloader = self._test_dataloader

    def _train_dataloader(self):
        return self.__get_dataloader(self.configs["train"])
    
    def _val_dataloader(self):
        return self.__get_dataloader(self.configs["validation"])

    def _test_dataloader(self):
        return self.__get_dataloader(self.configs["test"])

    def __get_dataloader(self, config) -> DataLoader:
        random_split = float(config["random_split"])
        if not (0 < random_split <= 1.0):
            Logger.error("random_split: (0.0, 1.0]")
            sys.exit(0)

        dataset = instantiate_from_config(config)
        if random_split != 1.0:
            Logger.info(f"RandomSplitDataset, random_split:{random_split:.2f}")
            dataset = RandomSplitDataset(dataset, random_split=random_split)

        return DataLoader(dataset=dataset, **config["data_loader"])


def single_to_batch(data_dict):
    batch = dict()
    for key, value in data_dict.items():
        if isinstance(value, torch.FloatTensor):
            batch[key] = value.unsqueeze(0)
        elif isinstance(value, int):
            batch[key] = torch.LongTensor(value).unsqueeze(0)
        elif isinstance(value, np.ndarray):
            batch[key] = torch.from_numpy(value).unsqueeze(0)
        else:
            Logger.error(f"key:{key}, {type(value)}")
    return batch



import os
import os.path as osp

import torch

from .config import get_log_dir, instantiate_from_config, load_config

__all__ = ["load_checkpoint", "get_best_checkpoint", "get_model"]


def load_checkpoint(model, checkpoint_file):
    state_dict = torch.load(checkpoint_file)["state_dict"]
    new_state_dict = dict()
    for key, value in state_dict.items():
        new_state_dict[key[len("model.") :]] = value
    model.load_state_dict(new_state_dict)
    return model


def get_best_checkpoint_from_folder(folder, select="min"):
    """获取最好的模型文件
    select: min, max
    """
    assert select in ["min", "max"]

    name_list = os.listdir(folder)
    if "last.ckpt" in name_list:
        name_list.remove("last.ckpt")

    best_score = None
    best_name = None
    for name in name_list:
        score = float(name.split("=")[-1].replace(".ckpt", ""))
        if best_score is None:
            best_score, best_name = score, name

        if select == "min" and score < best_score:
            best_score, best_name = score, name
        elif select == "max" and score > best_score:
            best_score, best_name = score, name

    if best_name is None:
        return None

    return osp.join(folder, best_name)


def get_best_checkpoint(config_filename, select="min"):
    checkpoint_folder = osp.join(get_log_dir(config_filename), "checkpoint")
    return get_best_checkpoint_from_folder(checkpoint_folder, select=select)


def get_model(config_filename):
    config = load_config(config_filename)
    model = instantiate_from_config(config.model)
    return model



"""
Image processing tools

Modified from open source projects:
(https://github.com/nkolot/GraphCMR/)
(https://github.com/open-mmlab/mmdetection)

"""

# copy from https://github.com/microsoft/MeshTransformer/tree/main/metro/utils/image_ops.py

import base64

import cv2
import numpy as np
import torch


def img_from_base64(imagestring):
    try:
        jpgbytestring = base64.b64decode(imagestring)
        nparr = np.frombuffer(jpgbytestring, np.uint8)
        r = cv2.imdecode(nparr, cv2.IMREAD_COLOR)
        return r
    except ValueError:
        return None


def img_to_base64(img):
    return base64.b64encode(cv2.imencode('.jpg', img)[1])


def myimrotate(img,
               angle,
               center=None,
               scale=1.0,
               border_value=0,
               auto_bound=False):
    if center is not None and auto_bound:
        raise ValueError('`auto_bound` conflicts with `center`')
    h, w = img.shape[:2]
    if center is None:
        center = ((w - 1) * 0.5, (h - 1) * 0.5)
    assert isinstance(center, tuple)

    matrix = cv2.getRotationMatrix2D(center, angle, scale)
    if auto_bound:
        cos = np.abs(matrix[0, 0])
        sin = np.abs(matrix[0, 1])
        new_w = h * sin + w * cos
        new_h = h * cos + w * sin
        matrix[0, 2] += (new_w - w) * 0.5
        matrix[1, 2] += (new_h - h) * 0.5
        w = int(np.round(new_w))
        h = int(np.round(new_h))
    rotated = cv2.warpAffine(img, matrix, (w, h), borderValue=border_value)
    return rotated


def myimresize(img, size, return_scale=False, interpolation='bilinear'):

    h, w = img.shape[:2]
    resized_img = cv2.resize(img, (size[0], size[1]),
                             interpolation=cv2.INTER_LINEAR)
    if not return_scale:
        return resized_img
    else:
        w_scale = size[0] / w
        h_scale = size[1] / h
        return resized_img, w_scale, h_scale


def get_transform(center, scale, res, rot=0):
    """Generate transformation matrix."""
    h = 200 * scale
    t = np.zeros((3, 3))
    t[0, 0] = float(res[1]) / h
    t[1, 1] = float(res[0]) / h
    t[0, 2] = res[1] * (-float(center[0]) / h + .5)
    t[1, 2] = res[0] * (-float(center[1]) / h + .5)
    t[2, 2] = 1
    if not rot == 0:
        rot = -rot  # To match direction of rotation from cropping
        rot_mat = np.zeros((3, 3))
        rot_rad = rot * np.pi / 180
        sn, cs = np.sin(rot_rad), np.cos(rot_rad)
        rot_mat[0, :2] = [cs, -sn]
        rot_mat[1, :2] = [sn, cs]
        rot_mat[2, 2] = 1
        # Need to rotate around center
        t_mat = np.eye(3)
        t_mat[0, 2] = -res[1] / 2
        t_mat[1, 2] = -res[0] / 2
        t_inv = t_mat.copy()
        t_inv[:2, 2] *= -1
        t = np.dot(t_inv, np.dot(rot_mat, np.dot(t_mat, t)))
    return t


def transform(pt, center, scale, res, invert=0, rot=0):
    """Transform pixel location to different reference."""
    t = get_transform(center, scale, res, rot=rot)
    if invert:
        # t = np.linalg.inv(t)
        t_torch = torch.from_numpy(t)
        t_torch = torch.inverse(t_torch)
        t = t_torch.numpy()
    new_pt = np.array([pt[0] - 1, pt[1] - 1, 1.]).T
    new_pt = np.dot(t, new_pt)
    return new_pt[:2].astype(int) + 1


def crop(img, center, scale, res, rot=0):
    """Crop image according to the supplied bounding box."""
    # Upper left point
    ul = np.array(transform([1, 1], center, scale, res, invert=1)) - 1
    # Bottom right point
    br = np.array(
        transform([res[0] + 1, res[1] + 1], center, scale, res, invert=1)) - 1
    # Padding so that when rotated proper amount of context is included
    pad = int(np.linalg.norm(br - ul) / 2 - float(br[1] - ul[1]) / 2)
    if not rot == 0:
        ul -= pad
        br += pad
    new_shape = [br[1] - ul[1], br[0] - ul[0]]
    if len(img.shape) > 2:
        new_shape += [img.shape[2]]
    new_img = np.zeros(new_shape)

    # Range to fill new array
    new_x = max(0, -ul[0]), min(br[0], len(img[0])) - ul[0]
    new_y = max(0, -ul[1]), min(br[1], len(img)) - ul[1]
    # Range to sample from original image
    old_x = max(0, ul[0]), min(len(img[0]), br[0])
    old_y = max(0, ul[1]), min(len(img), br[1])

    new_img[new_y[0]:new_y[1], new_x[0]:new_x[1]] = img[old_y[0]:old_y[1],
                                                        old_x[0]:old_x[1]]
    if not rot == 0:
        # Remove padding
        # new_img = scipy.misc.imrotate(new_img, rot)
        new_img = myimrotate(new_img, rot)
        new_img = new_img[pad:-pad, pad:-pad]

    # new_img = scipy.misc.imresize(new_img, res)
    new_img = myimresize(new_img, [res[0], res[1]])
    return new_img


def uncrop(img, center, scale, orig_shape, rot=0, is_rgb=True):
    """'Undo' the image cropping/resizing.
    This function is used when evaluating mask/part segmentation.
    """
    res = img.shape[:2]
    # Upper left point
    ul = np.array(transform([1, 1], center, scale, res, invert=1)) - 1
    # Bottom right point
    br = np.array(
        transform([res[0] + 1, res[1] + 1], center, scale, res, invert=1)) - 1
    # size of cropped image
    crop_shape = [br[1] - ul[1], br[0] - ul[0]]

    new_shape = [br[1] - ul[1], br[0] - ul[0]]
    if len(img.shape) > 2:
        new_shape += [img.shape[2]]
    new_img = np.zeros(orig_shape, dtype=np.uint8)
    # Range to fill new array
    new_x = max(0, -ul[0]), min(br[0], orig_shape[1]) - ul[0]
    new_y = max(0, -ul[1]), min(br[1], orig_shape[0]) - ul[1]
    # Range to sample from original image
    old_x = max(0, ul[0]), min(orig_shape[1], br[0])
    old_y = max(0, ul[1]), min(orig_shape[0], br[1])
    # img = scipy.misc.imresize(img, crop_shape, interp='nearest')
    img = myimresize(img, [crop_shape[0], crop_shape[1]])
    new_img[old_y[0]:old_y[1], old_x[0]:old_x[1]] = img[new_y[0]:new_y[1],
                                                        new_x[0]:new_x[1]]
    return new_img


def rot_aa(aa, rot):
    """Rotate axis angle parameters."""
    # pose parameters
    R = np.array([[np.cos(np.deg2rad(-rot)), -np.sin(np.deg2rad(-rot)), 0],
                  [np.sin(np.deg2rad(-rot)),
                   np.cos(np.deg2rad(-rot)), 0], [0, 0, 1]])
    # find the rotation of the body in camera frame
    per_rdg, _ = cv2.Rodrigues(aa)
    # apply the global rotation to the global orientation
    resrot, _ = cv2.Rodrigues(np.dot(R, per_rdg))
    aa = (resrot.T)[0]
    return aa


def flip_img(img):
    """Flip rgb images or masks.
    channels come last, e.g. (256,256,3).
    """
    img = np.fliplr(img)
    return img


def flip_kp(kp):
    """Flip keypoints."""
    flipped_parts = [
        5, 4, 3, 2, 1, 0, 11, 10, 9, 8, 7, 6, 12, 13, 14, 15, 16, 17, 18, 19,
        21, 20, 23, 22
    ]
    kp = kp[flipped_parts]
    kp[:, 0] = -kp[:, 0]
    return kp

def flip_kp_14(kp):
    """Flip keypoints."""
    # 0           1             2               3           4           5           6          
    # 'R_Ankle', 'R_Knee',      'R_Hip',        'L_Hip',    'L_Knee',   'L_Ankle',  'R_Wrist', 
    # 7           8             9               10          11          12          13   
    # 'R_Elbow', 'R_Shoulder',  'L_Shoulder',   'L_Elbow',  'L_Wrist',  'Neck',     'Head'
    flipped_parts = [
        5, 4, 3, 2, 1, 0, 11, 10, 9, 8, 7, 6, 12, 13
    ]
    kp = kp[flipped_parts]
    kp[:, 0] = -kp[:, 0]
    return kp

def flip_pose(pose):
    """Flip pose.
    The flipping is based on SMPL parameters.
    """
    flippedParts = [
        0, 1, 2, 6, 7, 8, 3, 4, 5, 9, 10, 11, 15, 16, 17, 12, 13, 14, 18, 19,
        20, 24, 25, 26, 21, 22, 23, 27, 28, 29, 33, 34, 35, 30, 31, 32, 36, 37,
        38, 42, 43, 44, 39, 40, 41, 45, 46, 47, 51, 52, 53, 48, 49, 50, 57, 58,
        59, 54, 55, 56, 63, 64, 65, 60, 61, 62, 69, 70, 71, 66, 67, 68
    ]
    pose = pose[flippedParts]
    # we also negate the second and the third dimension of the axis-angle
    pose[1::3] = -pose[1::3]
    pose[2::3] = -pose[2::3]
    return pose


def flip_aa(aa):
    """Flip axis-angle representation.
    We negate the second and the third dimension of the axis-angle.
    """
    aa[1] = -aa[1]
    aa[2] = -aa[2]
    return aa





class BaseTrackObj(object):
    def __init__(self) -> None:
        self._name = None
        self._time = None
        self._abs_time = None
        self._step = None

    def set_name(self, name):
        self._name = name

    def data(self):
        raise NotImplementedError("not implemented!")

    def show(self):
        raise NotImplementedError("not implemented!")



from ..tools.logger import Logger

def run_profile():
    text = """
        How to use:
        1. pip install snakeviz
        2. python -m cProfile -o out.prof train.py
        3. snakeviz out.prof
    """
    Logger.info(text)



import time


def get_time_str(time_float=-1):
    """获取当前时间字符串
    格式: 2021.05.17 12:20:20
    """

    if time_float == -1:
        time_float = time.time()
        assert isinstance(time_float, float)
    time_local = time.localtime(time_float)
    return time.strftime("%Y.%m.%d %H:%M:%S", time_local)

def get_time_folder():
    """获取以时间命名的文件夹名称 20210517_122020"""

    time_local = time.localtime(time.time())
    dt = time.strftime("%Y%m%d_%H%M%S", time_local)
    return dt

def get_date_str(time_float=-1):
    """ 获取日期 2021.05.17"""
    return get_time_str(time_float).split(' ')[0]



import os.path as osp

import numpy as np
import smplx
import torch
import torch.nn as nn
from smplx.lbs import vertices2joints

from ..keypoints.human import JOINTS_IDX as SMPL_J24


class BaseSMPL(smplx.SMPL):
    def __init__(self, model_path, gender="neutral"):
        super().__init__(
            model_path=osp.join(model_path, "smpl"),
            gender=gender,
            create_betas=False,
            create_body_pose=False,
            create_global_orient=False,
            create_transl=False,
            use_hands=False,
            use_feet_keypoints=False,
        )


class SMPL(nn.Module):
    def __init__(
        self,
        model_path,
    ):
        super().__init__()

        self._smpl_dict = nn.ModuleDict(
            {
                "neutral": BaseSMPL(model_path, gender="neutral"),
                "male": BaseSMPL(model_path, gender="male"),
                "female": BaseSMPL(model_path, gender="female"),
            }
        )
        self.faces = self._smpl_dict["neutral"].faces

        J_regressor_extra = torch.from_numpy(
            np.load(osp.join(model_path, "J_regressor_extra.npy"))
        ).float()
        self.register_buffer("J_regressor_extra", J_regressor_extra)

        J_regressor_h36m_correct = torch.from_numpy(
            np.load(osp.join(model_path, "J_regressor_h36m_correct.npy"))
        ).float()
        self.register_buffer("J_regressor_h36m_correct", J_regressor_h36m_correct)

    def forward(self, betas, pose, gender="neutral", return_dict=False):
        gender = gender.lower()
        if gender == "f":
            gender = "female"
        elif gender == "m":
            gender = "male"

        output = self._smpl_dict[gender](
            betas, body_pose=pose[:, 3:], global_orient=pose[:, :3]
        )
        # vertices = output["vertices"]
        # joints = output["joints"]
        if return_dict:
            return output
        return output["vertices"]

    def get_h36m_joints(self, vertices):
        """h36m: 17 joints"""
        return vertices2joints(self.J_regressor_h36m_correct, vertices)

    def get_joints(self, vertices, gender="neutral"):
        """get 24 joints"""
        gender = gender.lower()
        if gender == "f":
            gender = "female"
        elif gender == "m":
            gender = "male"

        joints = vertices2joints(self._smpl_dict[gender].J_regressor, vertices)
        joints_extra = vertices2joints(self.J_regressor_extra, vertices)
        # 38 joints
        joints = torch.cat((joints, joints_extra), dim=1)
        # 24 joints
        joints = joints[:, SMPL_J24]
        return joints



import pickle
from typing import List

from .track_base import BaseTrackObj


def load_trackfile(trackfile) -> List[BaseTrackObj]:
    obj_list = []
    with open(trackfile, "rb") as fr:
        while True:
            try:
                obj = pickle.load(fr)
                obj_list.append(obj)
            except EOFError:
                break
    return obj_list


def analyzer(trackfile):
    obj_list = load_trackfile(trackfile)
    for obj in obj_list:
        obj.show()
